/*
 * File:	Arrange.java
 * author:	Haitao Liu (haitaol@cs.arizona.edu)
 *
 * The tauXSchema toolset, Version 1.0
 *
 * Copyright (c) 2002-2003 
 * The Department of Computer Science of the University of Arizona.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by 
 *       the Department of Computer Science of the University of Arizona."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear. 
 *
 * 4. The names "tauXSchema" or "TAU (Temporal Access for Users)" 
 *    must not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact tau@cs.arizona.edu.
 *
 * 5. Products derived from this software may not be called "tauXSchema",
 *    nor may "tauXSchema" or "TAU (Temporal Access for Users)" appear in 
 *    their name, without prior written permission of the Department of 
 *    Computer Science of the University of Arizona.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE Department of Computer Science of 
 * the University of Arizona OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This toolset contains useful tools to manage temporal XML document.
 * For more information about tXSchema and temporal XML document, 
 * visit the homepage of TAU Project <http://www.cs.arizona.edu/tau/>.
 */

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.apache.xpath.*;
import java.util.*;
import org.apache.xerces.parsers.*;
import org.apache.xerces.dom3.*;
import org.w3c.dom.ls.*;
import java.io.*;
import org.w3c.dom.traversal.*;
import org.apache.xerces.parsers.*;
import org.xml.sax.*;
import org.apache.xpath.*;

/* 
 *	adjust the ordering of children of the elements whose definitions are changed
 *	in representational schema to address the xs:all to xs:sequenced problem
 */
public class Arrange
{
	static DOMParser parser;
	static DOMWriter writer;
	
	static Hashtable def_table;
	
	public static void main(String[] args) throws Exception{
		init();
		
		Document doc, schema, phy_anno;
		
		parser.parse(new InputSource(args[0]));
		doc = parser.getDocument();
		
		parser.parse(new InputSource(args[1]));
		schema = parser.getDocument();
		
		parser.parse(new InputSource(args[2]));
		phy_anno = parser.getDocument();
		
		Document modi_doc = do_arrange(doc, schema, phy_anno);
		
		writer.writeNode(new FileOutputStream(args[0]), modi_doc);
	}
	
	static void init() throws Exception{
		parser = new DOMParser();
		
		System.setProperty(DOMImplementationRegistry.PROPERTY,
    		"org.apache.xerces.dom.DOMImplementationSourceImpl");
		DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();

		DOMImplementationLS impl =
			(DOMImplementationLS)registry.getDOMImplementation("LS-Load");;

		writer = impl.createDOMWriter();
	}
	
	/* rearrange children according to ordering given in definition */
	public static void arrange_children(Element parent, Element def) throws Exception{
		ArrayList child_order = new ArrayList();
		
		/* get element's type declaration	*/
		Node n = def.getFirstChild();
		while (!(n instanceof Element) && !n.getNodeName().equals("xs:complexType"))
			n = n.getNextSibling();
			
		/* get element's compositor declaration */
		n = n.getFirstChild();
		while (!(n instanceof Element))
			n = n.getNextSibling();
		
		/* get the ordering of children */
		n = n.getFirstChild();
		while (n != null){
			if (n.getNodeName().equals("xs:element"))
				child_order.add(n);
				
			n = n.getNextSibling();
		}
		
		Node anchor = parent.getFirstChild();
		for (int i = 0; i < child_order.size(); i++){
			String c_name = ((Element)child_order.get(i)).getAttribute("name");
			if (c_name.length() == 0)
				c_name = ((Element)child_order.get(i)).getAttribute("ref");			
			
			NodeList cl = XPathAPI.selectNodeList(parent, c_name);
			
			ArrayList cl_array = new ArrayList();
			
			for (int k = 0; k < cl.getLength(); k++)
				cl_array.add(cl.item(k));
			
			for (int k = 0; k < cl_array.size(); k++){
				Node new_n = parent.insertBefore((Node)cl_array.get(k), anchor);
				
				anchor = new_n.getNextSibling();
			}
		}
	}
	
	public static Document do_arrange(Document doc, Document schema, Document phy_anno) throws Exception{
		/* get top level definitions */
		get_defs(schema.getDocumentElement());		
		
		NodeList vt = phy_anno.getElementsByTagName("stampPosition");
		
		for (int i = 0; i < vt.getLength(); i++){			
			Element vt_item = (Element)vt.item(i);	
			
			String path = vt_item.getAttribute("target");
			
			path = path.substring(1);
			
			int first_slash, last_slash;
			
			first_slash = path.indexOf("/");
			if (first_slash < 0)	// root has no parent
				continue;
				
			path = path.substring(first_slash + 1);	// get relative path to root
							
			last_slash = path.lastIndexOf("/");			
			if (last_slash < 0){
				// the timestampable element is child of root				
				Element parent = doc.getDocumentElement();
				
				Element def = find_def(parent);
				
				arrange_children(parent, def);
			}
			else{
				// get path to the parent of timestampable element
				path = path.substring(0, last_slash);
				
				NodeList parent_list = XPathAPI.selectNodeList(doc.getDocumentElement(), path);
				
				if (parent_list.getLength() == 0)
					continue;
				
				Element def = find_def((Element)parent_list.item(0));
								
				for (int j = 0; j < parent_list.getLength(); j++){
					Element parent = (Element)parent_list.item(j);	
					
					arrange_children(parent, def);
				}
			}
		}
		
		return doc;
	}
	
	static void get_defs(Element sch_root) throws Exception{
		def_table = new Hashtable();
		
    	NodeList nl = sch_root.getChildNodes();
		
		for (int i = 0; i < nl.getLength(); i++){
			Node n = nl.item(i);
			
			if (!n.getNodeName().equals("xs:element"))
				continue;
				
			Element def = (Element)n;
			
			def_table.put(def.getAttribute("name"), def);
		}
		
		return;
	}
	
	/* find the definition of an element in schema */
	static Element find_def(Element e){
		Element def;
		
		Element root_def = (Element)def_table.get(e.getOwnerDocument().getDocumentElement().getNodeName());
		
		if (e.equals(e.getOwnerDocument().getDocumentElement()))
			return root_def;
		
		// get the names of nodes on the path to e
		ArrayList node_name = new ArrayList();
		node_name.add(e.getNodeName());
		
		Node parent = e.getParentNode();
		while (parent instanceof Element){
			node_name.add(0, parent.getNodeName());
			parent = parent.getParentNode();
		}
		
		TreeWalker tw_def = ((DocumentTraversal)root_def.getOwnerDocument()).createTreeWalker(root_def,
														NodeFilter.SHOW_ELEMENT,
														null,
														false);
		ArrayList nodes_past = new ArrayList();
		
		int posi = 1;
		
		Node n = tw_def.firstChild();
		while (true){
			if (n == null){
				if (nodes_past.size() == 0)
					return null;
				
				n = (Node)nodes_past.remove(nodes_past.size()-1);
				
				tw_def.setCurrentNode(n);
				n = tw_def.nextSibling();	
				
				continue;
			}
			
			if (!n.getNodeName().equals("xs:element")){
				nodes_past.add(n);
				
				tw_def.setCurrentNode(n);
				n = tw_def.firstChild();
				
				continue;
			}
			
			def = (Element)n;
			
			if (def.getAttribute("name").equals(node_name.get(posi)) ||
					def.getAttribute("ref").equals(node_name.get(posi))){
				if (def.getAttribute("ref").length() > 0)
					def = (Element)def_table.get(def.getAttribute("ref"));
					
				posi++;
				
				if (posi >= node_name.size())
					break;
				
				tw_def.setCurrentNode(def);
				n = tw_def.firstChild();
			}
			else{
				tw_def.setCurrentNode(n);
				n = tw_def.nextSibling();	
			}
		}										
					
		return def;
	}	
}










