/*
 * File:	Squash.java
 * author:	Haitao Liu (haitaol@cs.arizona.edu)
 *
 * The tauXSchema toolset, Version 1.0
 *
 * Copyright (c) 2002-2003 
 * The Department of Computer Science of the University of Arizona.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by 
 *       the Department of Computer Science of the University of Arizona."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear. 
 *
 * 4. The names "tauXSchema" or "TAU (Temporal Access for Users)" 
 *    must not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact tau@cs.arizona.edu.
 *
 * 5. Products derived from this software may not be called "tauXSchema",
 *    nor may "tauXSchema" or "TAU (Temporal Access for Users)" appear in 
 *    their name, without prior written permission of the Department of 
 *    Computer Science of the University of Arizona.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE Department of Computer Science of 
 * the University of Arizona OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This toolset contains useful tools to manage temporal XML document.
 * For more information about tXSchema and temporal XML document, 
 * visit the homepage of TAU Project <http://www.cs.arizona.edu/tau/>.
 */

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.apache.xpath.*;
import java.util.*;
import org.apache.xerces.dom3.*;
import org.apache.xerces.parsers.*;
import org.w3c.dom.ls.*;
import java.io.*;
import org.w3c.dom.traversal.*;
import org.apache.xerces.parsers.*;
import org.xml.sax.*;
import java.text.*;

public class Squash
{
	static final String FOREVER = "9999-12-31";
	
	static Document cur_doc = null;
	
	static String begin_time, cur_time, next_time;
	
	static DOMParser parser;
	static DOMWriter writer;
	
	static Document log_anno = null, phy_anno = null;
	static AnnoParser anno_parser;
	
	static ArrayList snapshot_list, ts_list;
	
	static SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
	
	static String txml_name;
	
	/* create parser and writer */
	static void init() throws Exception{
		parser = new DOMParser();
		
        System.setProperty(DOMImplementationRegistry.PROPERTY,
    		"org.apache.xerces.dom.DOMImplementationSourceImpl");
		DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();

		DOMImplementationLS impl =
			(DOMImplementationLS)registry.getDOMImplementation("LS-Load");

		writer = impl.createDOMWriter();
	}
	
	/* get Squash parameters from configuration file */
	static void squash_by_conf(String conf_file) throws Exception{
		parser.parse(new InputSource(conf_file));
		
		txml_name = "temp_doc.xml";
		
		Element conf_root = parser.getDocument().getDocumentElement();		
		
		/* read logical and physical annotation */
		String log_schema, phy_schema;
		log_schema = conf_root.getAttribute("log_anno");
		phy_schema = conf_root.getAttribute("phy_anno");
		
		if (log_schema.length() > 0){
			parser.parse(new InputSource(log_schema));
			log_anno = parser.getDocument();	
		}
		
		if (phy_schema.length() > 0){
			parser.parse(new InputSource(phy_schema));
			phy_anno = parser.getDocument();
			
			if (log_anno == null){
				NodeList nl = phy_anno.getElementsByTagName("temporalAnnotatations");
				
				if (nl.getLength() > 0){
					log_schema = ((Element)nl.item(0)).getAttribute("location");
					parser.parse(new InputSource(log_schema));
					log_anno = parser.getDocument();
				}				
			}
		}

		/* retrieve file names of snapshots and their timestamps from configuration file */
		snapshot_list = new ArrayList();
		ts_list = new ArrayList();
		
		NodeList snap_list = conf_root.getElementsByTagName("snapshot");
		for (int i = 0; i < snap_list.getLength(); i++){
			Element n = (Element)snap_list.item(i);
			
			String name, ts;
			
			name = n.getAttribute("file");
			ts = n.getAttribute("time");
			
			int posi = -1;
			for (int j = 0; j < ts_list.size(); j++){
				String ts1 = (String)ts_list.get(j);
				
				if (sdf.parse(ts1).compareTo(sdf.parse(ts)) > 0){
					posi = j;
					break;	
				}	
			}
			
			if (posi == -1){
				parser.parse(new InputSource(name));				
				snapshot_list.add(parser.getDocument());
				
				ts_list.add(ts);
			}
			else{
				parser.parse(new InputSource(name));				
				snapshot_list.add(posi, parser.getDocument());
				
				ts_list.add(posi, ts);
			}
		}
		
		squash();
	}
	
	/* get Squash parameters from command arguments */
	static void squash_by_param(String[] args) throws Exception {
		String prefix = args[0];
		
		txml_name = "temp_" + prefix.substring(prefix.lastIndexOf("/") + 1) + ".xml";
			
		int begin_idx = Integer.parseInt(args[1]);
		int end_idx = Integer.parseInt(args[2]);
		
		for (int i = 3; i < args.length; i++){
			parser.parse(new InputSource(args[i]));
			
			Document doc = parser.getDocument();
			
			if (doc.getDocumentElement().getNodeName().equals("temporalAnnotatedSchema"))
				log_anno = doc;
			else if (doc.getDocumentElement().getNodeName().equals("physicalAnnotatedSchema")){
				phy_anno = doc;
				
				if (log_anno == null){
					NodeList nl = phy_anno.getElementsByTagName("temporalAnnotatations");
					if (nl.getLength() > 0){
						String location = ((Element)nl.item(0)).getAttribute("location");	
						
						parser.parse(new InputSource(location));
						log_anno = parser.getDocument();
					}
				}
			}
			else{
				System.out.println("Unrecognized parameter -- " + args[i]);
				System.exit(0);
			}
		}
		
		snapshot_list = new ArrayList();
		ts_list = new ArrayList();
		
		for (int i = begin_idx; i <= end_idx; i++){			
			String fname = prefix + i + ".xml";			
			
			parser.parse(new InputSource(new FileInputStream(fname)));
			Document snap_doc = parser.getDocument();
			
			snapshot_list.add(snap_doc);
			
			Element snap_root = snap_doc.getDocumentElement();
			
			String ts = snap_root.getAttribute("cur_date");
			ts_list.add(ts);
		}		
		
		squash();
	}
		
	public static void main(String[] args) throws Exception{
		if (args.length == 0){
			System.out.println("Usage: java Squash [config_file | prefix starting_index ending_index [logcial_schema] [physical_schema]");						
			System.exit(0);
		}
		
		long begin_tm = System.currentTimeMillis();		
		
		init();
		
		if (args.length == 1)
			squash_by_conf(args[0]);
		else			
			squash_by_param(args);		
				
		long end_tm = System.currentTimeMillis();
		
		System.out.println("time spent: " + (end_tm - begin_tm) + " milliseconds.");
	}		
	
	static void generate_tv_root_txml() throws Exception {
		Element cur_root = ((Document)snapshot_list.get(0)).getDocumentElement();
		cur_time = cur_root.getAttribute("cur_date");	
		
		String begin_time = cur_time;	
		
		// create timevarying root document
		Document tv_root_doc = 
				cur_root.getOwnerDocument().getImplementation().createDocument(null, "timevarying-root", null);
		Element tv_root = tv_root_doc.getDocumentElement();
		
		Element ts = tv_root_doc.createElement("timestamp");			
		
		String last_verid = cur_root.getAttribute("cur_verid");
		
		/* integerate snapshots in time order */
		for (int i = 1; i < snapshot_list.size(); i++){			
			System.out.println("Integrate snapshot " + i);			
			
			Element next_root = ((Document)snapshot_list.get(i)).getDocumentElement();				
			next_time = next_root.getAttribute("cur_date");
			
			last_verid = next_root.getAttribute("cur_verid");
								
			cur_root.removeAttribute("cur_verid");
			cur_root.removeAttribute("cur_date");
			
			Element cur_root1 = (Element)tv_root_doc.importNode(cur_root, true);
			
			ts.setAttribute("vtbegin", cur_time);
			ts.setAttribute("vtend", next_time);
			
			cur_root1.insertBefore(ts.cloneNode(true), cur_root1.getFirstChild());
			
			tv_root.appendChild(cur_root1);
			
			cur_root = next_root;
			cur_time = next_time;
		}
		
		/* integrate last snapshot */
		Element cur_root1 = (Element)tv_root_doc.importNode(cur_root, true);
		
		cur_root1.removeAttribute("cur_verid");
		cur_root1.removeAttribute("cur_date");
			
		ts.setAttribute("vtbegin", cur_time);
		ts.setAttribute("vtend", FOREVER);
		
		cur_root1.insertBefore(ts.cloneNode(true), cur_root1.getFirstChild());
		
		tv_root.appendChild(cur_root1);
		
		tv_root.setAttribute("cur_verid", last_verid);
		tv_root.setAttribute("cur_date", begin_time);
		
		writer.writeNode(new FileOutputStream(txml_name), tv_root_doc);
	}	
	
	static void squash() throws Exception{		
		if ((phy_anno != null ) && (log_anno == null)){
			System.out.println("Logical annotation is missing while physical annotation is present!");
			System.exit(0);	
		}
		
		if (phy_anno == null){
			generate_tv_root_txml();	
			
			return;
		}
		
		anno_parser = new AnnoParser(log_anno, phy_anno);
		
		String last_verid = "";
		
		/* get the first snapshot */
		cur_doc = (Document)snapshot_list.get(0);
		
		last_verid = cur_doc.getDocumentElement().getAttribute("cur_verid");
		
		cur_time = (String)ts_list.get(0);
		
		begin_time = cur_time;
		
		/* remove these attributes so that they are not treated as timevarying attributes during merging */
		cur_doc.getDocumentElement().removeAttribute("cur_verid");		
		cur_doc.getDocumentElement().removeAttribute("cur_date");		
		
		for (int i = 1; i < snapshot_list.size(); i++){
			System.out.println("Integrate snapshot " + i);			
			
			Document next_doc = (Document)snapshot_list.get(i);
			
			Element next_root = (Element)cur_doc.importNode(next_doc.getDocumentElement(), true);			
			
			next_time = (String)ts_list.get(i);
			
			last_verid = next_root.getAttribute("cur_verid");
			
			next_root.removeAttribute("cur_verid");
			next_root.removeAttribute("cur_date");
								
			merge_non_ts_nodes(cur_doc.getDocumentElement(), next_root);
		}		
		
		cur_doc.getDocumentElement().setAttribute("cur_date", begin_time);
		cur_doc.getDocumentElement().setAttribute("cur_verid", last_verid);
		
		cur_doc.getDocumentElement().removeAttribute("xmlns:xsi");
		
		cur_doc.getDocumentElement().removeAttribute("xsi:noNamespaceSchemaLocation");
		
		writer.writeNode(new FileOutputStream(txml_name), cur_doc);
	}	
	
	/* 
	 *	compare two elements' tv-attrs that can't be timestamped
	 */
	static boolean equal_nonts_tvattr(Element e1, Element e2){
		NamedNodeMap map1, map2;
		
		map1 = e1.getAttributes();		
		map2 = e2.getAttributes();
		
		/* put non-timestampable attributes of e1 in arraylist */
		ArrayList tv_attr1 = new ArrayList();			
		
		for (int i = 0; i < map1.getLength(); i++){
			Attr attr1 = (Attr)map1.item(i);
			
			String attr_name = attr1.getNodeName();
			
			if (!anno_parser.hasTVAttr(e1, attr_name))
				continue;
				
			if (anno_parser.hasTimestampableTVAttr(e1, attr_name))
				continue;
				
			tv_attr1.add(attr1);	
		}
		
		/* put values of non-timestampable attributes of e2 in hashtable, key is attribute name */
		Hashtable tv_attr2 = new Hashtable();
		for (int i = 0; i < map2.getLength(); i++){
			Attr attr2 = (Attr)map2.item(i);
			
			String attr_name = attr2.getNodeName();
			
			if (!anno_parser.hasTVAttr(e2, attr_name))
				continue;
				
			if (anno_parser.hasTimestampableTVAttr(e2, attr_name))
				continue;
				
			tv_attr2.put(attr_name, attr2.getNodeValue());	
		}
		
		/* compare size */
		if (tv_attr1.size() != tv_attr2.size())
			return false;
			
		/* compare value */
		for (int i = 0; i < tv_attr1.size(); i++){
			Attr ta = (Attr)tv_attr1.get(i);
			
			String name = ta.getNodeName();
			String value = ta.getNodeValue();
			
			if (!value.equals(tv_attr2.get(name)))
				return false;
		}
		
		return true;
	}
	
	/* processe timestampable timevarying attributes */
	static void process_tv_attr(Element cur_e, Element next_e){
		NamedNodeMap cur_map, next_map;
		
		cur_map = cur_e.getAttributes();
		next_map = next_e.getAttributes();
		
		Hashtable cur_reg_attrs = new Hashtable();
		for (int i = 0; i < cur_map.getLength(); i++)
			cur_reg_attrs.put(cur_map.item(i).getNodeName(), cur_map.item(i));
		
		Hashtable cur_tv_attrs = new Hashtable();
		
		TreeWalker tw = ((DocumentTraversal)cur_e.getOwnerDocument()).createTreeWalker(cur_e,
    															NodeFilter.SHOW_ELEMENT,
    															new TS_TVAttrFilterIn(),
    															false);
		Element ta = (Element)tw.firstChild();
		while (ta != null){
			if (ta.getAttribute("vtend").equals(FOREVER))
				cur_tv_attrs.put(ta.getAttribute("name"), ta);
			
			tw.setCurrentNode(ta);
			ta = (Element)tw.nextSibling();
		}
		
		/* compare values of attributes in next snapshot with values in current snapshot */
		for (int i = 0; i < next_map.getLength(); i++){
			Attr next_attr = (Attr)next_map.item(i);
			
			String attr_name = next_attr.getNodeName();
			String next_value = next_attr.getNodeValue();
			
			/* search in regular attributes in current snapshot */
			Attr old_attr = (Attr)cur_reg_attrs.get(attr_name);
			
			if (old_attr == null){
				/* search in timevarying attribute in current snapshot */
				Element tv = (Element)cur_tv_attrs.get(attr_name);				
				
				if (tv == null){
					tv = create_tv_attr(attr_name, next_value, next_time, FOREVER);
					cur_e.insertBefore(tv, cur_e.getFirstChild());
				}
				else{
					String old_value = tv.getAttribute("value");
					
					if (!old_value.equals(next_value)){
						tv.setAttribute("vtend", next_time);
						
						Element next_tv = create_tv_attr(attr_name, next_value, next_time, FOREVER);
						cur_e.insertBefore(next_tv, tv.getNextSibling());
					}
					
					cur_tv_attrs.remove(attr_name);
				}
			}
			else{
				cur_reg_attrs.remove(attr_name);
				
				String old_value = old_attr.getNodeValue();
				
				if (!old_value.equals(next_value)){
					cur_e.removeAttributeNode(old_attr);
					
					Element old_tv, next_tv;
					
					old_tv = create_tv_attr(attr_name, old_value, begin_time, next_time);
					next_tv = create_tv_attr(attr_name, next_value, next_time, FOREVER);
					
					old_tv = (Element)cur_e.insertBefore(old_tv, cur_e.getFirstChild());
					cur_e.insertBefore(next_tv, old_tv.getNextSibling());
				}
			}
		}
		
		/* timestamp left current attributes that don't appear in new snapshot */
		
		Enumeration enum = cur_reg_attrs.keys();
		while (enum.hasMoreElements()){
			Attr old_attr = (Attr)cur_reg_attrs.get(enum.nextElement());	
			
			Element old_tv;
					
			old_tv = create_tv_attr(old_attr.getNodeName(), 
									old_attr.getNodeValue(),
									begin_time, 
									next_time);
			
			cur_e.insertBefore(old_tv, cur_e.getFirstChild());
			
			cur_e.removeAttribute(old_attr.getNodeName());
		}
		
		enum = cur_tv_attrs.elements();
		while (enum.hasMoreElements()){
			Element tv = (Element)enum.nextElement();	
			tv.setAttribute("vtend", next_time);
		}
	}
	
	/* merge non-timestampable nodes */
	static void merge_non_ts_nodes(Element cur_e, Element next_e){
		ArrayList cur_non_ts_children, cur_ts_children;
		cur_non_ts_children = new ArrayList(); 		
		cur_ts_children = new ArrayList();
		
		cur_non_ts_children = get_non_ts_children(cur_e);		
		cur_ts_children = get_ts_children(cur_e);		
				
		ArrayList next_non_ts_children, next_ts_children;
		next_non_ts_children = new ArrayList(); 		
		next_ts_children = new ArrayList();
		
		next_non_ts_children = get_non_ts_children(next_e);
		next_ts_children = get_ts_children(next_e);
		
		if (cur_non_ts_children.size() != next_non_ts_children.size()){
			System.out.println("Something is wrong: try again after	removing redundant blanks in documents!");
			
			System.exit(0);
		}
		
		for (int i = 0; i < next_non_ts_children.size(); i++){
			Node cur_non_ts_c = (Node)cur_non_ts_children.get(i);
			Node next_non_ts_c = (Node)next_non_ts_children.get(i);
			
			if (next_non_ts_c instanceof Element)	
				merge_non_ts_nodes((Element)cur_non_ts_c, (Element)next_non_ts_c);
		}
		
		for (int i = 0; i < next_ts_children.size(); i++){
			Element next_ts_c = (Element)next_ts_children.get(i);
			
			Element cur_ts_c = search_by_verid(cur_ts_children,
										next_ts_c.getAttribute("verid"));
										
			if (cur_ts_c == null){
				timestamp_next(next_ts_c, next_time);
				
				cur_e.appendChild(next_ts_c);
				
				continue;	
			}
			
			cur_ts_children.remove(cur_ts_c);
			
			merge_ts_nodes(cur_ts_c, next_ts_c);
		}
		
		// the left children are timestamped because they don't appear in the new snapshot
		for (int i = 0; i < cur_ts_children.size(); i++)
			timestamp_cur((Element)cur_ts_children.get(i), next_time);
			
		// handle time-varying attribute
		process_tv_attr(cur_e, next_e);
	}
	
	static void merge_ts_nodes(Element cur_e, Element next_e){
		/*	
		 *	compare non-ts decadents, if they are different, 
		 *	make spearate copies of current and next element
		 */
		if (equal_non_ts_parts(cur_e, next_e) == false){			
			timestamp_cur(cur_e, next_time);
			
			timestamp_next(next_e, next_time);
			
			cur_e.getParentNode().insertBefore(next_e, cur_e.getNextSibling());
			
			return;
		}
		
		ArrayList cur_non_ts_children, cur_ts_children;
		cur_non_ts_children = new ArrayList(); 		
		cur_ts_children = new ArrayList();
		
		cur_non_ts_children = get_non_ts_children(cur_e);
		cur_ts_children = get_ts_children(cur_e);		
		
		ArrayList next_non_ts_children, next_ts_children;
		next_non_ts_children = new ArrayList(); 		
		next_ts_children = new ArrayList();
		
		next_non_ts_children = get_non_ts_children(next_e);
		next_ts_children = get_ts_children(next_e);
		
		// non-ts children should match by now, 		
		for (int i = 0; i < next_non_ts_children.size(); i++){
			Node cur_non_ts_c = (Node)cur_non_ts_children.get(i);	
			Node next_non_ts_c = (Node)next_non_ts_children.get(i);
			
			if (next_non_ts_c instanceof Element)
				merge_non_ts_nodes((Element)cur_non_ts_c, (Element)next_non_ts_c);
		}
		
		// merge timestampable children
		for (int i = 0; i < next_ts_children.size(); i++){
			Element next_ts_c = (Element)next_ts_children.get(i);
			
			Element cur_ts_c = search_by_verid(cur_ts_children,
										next_ts_c.getAttribute("verid"));
										
			if (cur_ts_c == null){
				timestamp_next(next_ts_c, next_time);
				
				cur_e.appendChild(next_ts_c);
				
				continue;	
			}
			
			cur_ts_children.remove(cur_ts_c);
			
			merge_ts_nodes(cur_ts_c, next_ts_c);
		}
		
		for (int i = 0; i < cur_ts_children.size(); i++)
			timestamp_cur((Element)cur_ts_children.get(i), next_time);		
			
		//Handle time-varying attribute
		process_tv_attr(cur_e, next_e);
	}
	
	static Element create_tv_attr(String name, String value, String vtbegin, String vtend){
		Element tv = cur_doc.createElement("timevarying-attr");
		
		tv.setAttribute("name", name);
		tv.setAttribute("value", value);
		tv.setAttribute("vtbegin", vtbegin);
		tv.setAttribute("vtend", vtend);
		
		return tv;
	}
	
	/* recursively compare non-timestampable decadents of two nodes */
	static boolean equal_non_ts_parts(Node n1, Node n2){
		if (n1.getNodeType() != n2.getNodeType())
			return false;
			
		if (!n1.getNodeName().equals(n2.getNodeName()))
			return false;
			
		if (!(n1 instanceof Element)){
			String v1 = n1.getNodeValue();
			String v2 = n2.getNodeValue();
			
			if ((v1 != null) && (!v1.equals(v2)))
				return false;
			
			if ((v1 == null) && (v2 != null))
				return false;
				
			return true;		
		}
		
		Element e1 = (Element)n1;
		Element e2 = (Element)n2;
		
		if (anno_parser.isTimeVarying(e1)){
			String e1_verid = ((Element)e1).getAttribute("verid");
			String e2_verid = ((Element)e2).getAttribute("verid");
			
			if (!e1_verid.equals(e2_verid))
				return false;
		}
		
		if (!equal_nonts_tvattr(e1, e2))
			return false;
		
		ArrayList children1 = get_non_ts_children(e1);
		ArrayList children2 = get_non_ts_children(e2);		
			
		if (children1.size() != children2.size())
			return false;
				
		for (int i = 0; i < children1.size(); i++){
			Node c1 = (Node)children1.get(i);
			Node c2 = (Node)children2.get(i);
			
			if (equal_non_ts_parts(c1, c2) == false)
				return false;
		}		
		
		return true;
	}	
	
	/* get the child having the specified verid in the given element */
	static Element search_by_verid(ArrayList nodes, String verid){
		for (int i = 0; i < nodes.size(); i++){
			Element e = (Element)nodes.get(i);
			
			String c_verid = e.getAttribute("verid");
			
			if ((c_verid != null) && (c_verid.equals(verid)))
				return e;
		}	
		
		return null;
	}
	
	static ArrayList get_non_ts_children(Element parent){
		ArrayList non_ts_children = new ArrayList();
		
		TreeWalker tw = ((DocumentTraversal)parent.getOwnerDocument()).createTreeWalker(parent,
    															NodeFilter.SHOW_ALL,
    															new TS_TAttrFilterOut(),
    															false);

		
		Node n = tw.firstChild();
		
		while (n != null){
			if (!(n instanceof Element)){				
				non_ts_children.add(n);
				
				tw.setCurrentNode(n);
				n = tw.nextSibling();
				
				continue;
			}
			
			if (!anno_parser.isTimestampableElement((Element)n))
				non_ts_children.add(n);
				
			tw.setCurrentNode(n);
			n = tw.nextSibling();
		}
		
		return non_ts_children;
	}
	
	static ArrayList get_ts_children(Element parent){
		ArrayList ts_children = new ArrayList();
		
		TreeWalker tw = ((DocumentTraversal)parent.getOwnerDocument()).createTreeWalker(parent,
    															NodeFilter.SHOW_ELEMENT,
    															new TS_TAttrFilterOut(),
    															false);
		Node n = tw.firstChild();
		
		while (n != null){
			Element e = (Element)n;
			
			if (anno_parser.isTimestampableElement(e)){
				TreeWalker tw1 = ((DocumentTraversal)cur_doc).createTreeWalker(e,
    															NodeFilter.SHOW_ELEMENT,
    															new TS_TSFilterIn(),
    															false);

    			Element ts = (Element)tw1.firstChild();
    			
    			// check if the node is valid at current instant
    			
    			if (ts == null)
    				ts_children.add(e);
    			else if (ts.getAttribute("vtend").equals(FOREVER))
    				ts_children.add(e);
			}
			
			tw.setCurrentNode(n);
			n = tw.nextSibling();
		}
		
		return ts_children;
	}
	
	static void timestamp_cur(Element e, String end){
		NodeList ts_nl = e.getElementsByTagName("timestamp");
		
		for (int i = 0; i < ts_nl.getLength(); i++){
			Element ts = (Element)ts_nl.item(i);
			
			if (ts.getAttribute("vtend").equals(FOREVER))				
				ts.setAttribute("vtend", end);
		}
			
		TreeWalker tw = ((DocumentTraversal)cur_doc).createTreeWalker(e,
															NodeFilter.SHOW_ELEMENT,
															new TS_TSFilterIn(),
															false);

		Element ts = (Element)tw.firstChild();
		
		if (ts == null){
			// find begin time from ancestor
			
			String begin;
			
			Node parent = e.getParentNode();
			
			while (parent instanceof Element){
				tw = ((DocumentTraversal)cur_doc).createTreeWalker((Element)parent,
															NodeFilter.SHOW_ELEMENT,
															new TS_TSFilterIn(),
															false);

				ts = (Element)tw.firstChild();
				
				if (ts != null)
					break;
					
				parent = parent.getParentNode();
			}
			
			if (ts == null)
				begin = begin_time;				
			else
				begin = ts.getAttribute("vtbegin");
			
			ts = cur_doc.createElement("timestamp");
				
			ts.setAttribute("vtbegin", begin);
			ts.setAttribute("vtend", end);
			
			e.insertBefore(ts, e.getFirstChild());
		}
		else		
			ts.setAttribute("vtend", end);		
	}
	
	static void timestamp_next(Element e, String begin){
		Element ts = cur_doc.createElement("timestamp");
				
		ts.setAttribute("vtbegin", begin);
		ts.setAttribute("vtend", FOREVER);
		
		e.insertBefore(ts, e.getFirstChild());		
	}
}
