/*
 * File:	TemporalValidator.java
 * author:	Haitao Liu (haitaol@cs.arizona.edu)
 *
 * The tauXSchema toolset, Version 1.0
 *
 * Copyright (c) 2002-2003 
 * The Department of Computer Science of the University of Arizona.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by 
 *       the Department of Computer Science of the University of Arizona."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear. 
 *
 * 4. The names "tauXSchema" or "TAU (Temporal Access for Users)" 
 *    must not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact tau@cs.arizona.edu.
 *
 * 5. Products derived from this software may not be called "tauXSchema",
 *    nor may "tauXSchema" or "TAU (Temporal Access for Users)" appear in 
 *    their name, without prior written permission of the Department of 
 *    Computer Science of the University of Arizona.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE Department of Computer Science of 
 * the University of Arizona OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This toolset contains useful tools to manage temporal XML document.
 * For more information about tXSchema and temporal XML document, 
 * visit the homepage of TAU Project <http://www.cs.arizona.edu/tau/>.
 */

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.apache.xpath.*;
import java.util.*;
import org.apache.xerces.dom3.*;
import org.apache.xerces.parsers.*;
import org.w3c.dom.ls.*;
import java.io.*;
import org.w3c.dom.traversal.*;
import org.apache.xerces.parsers.*;
import org.xml.sax.*;
import java.text.*;

public class TemporalValidator
{
	static final String FOREVER = "9999-12-31";
	
	static SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
	
	/* TreeWalker for finding timestamp or timevarying-attr */
	static TreeWalker ts_tw, tvattr_tw;
	
	static AnnoParser anno_parser;
	
	static DOMParser parser;
	
	/* store the top-level element definitions in non-temporal schema */
	static Hashtable def_table;
	
	/* temporal XML document */
	static Document doc;
	static Element doc_root = null;
	
	/* file names of temporal XML, non-temporal schema, logical and physical annotation and representataional schema */
	static String txml = null, nt_schema = null, log_schema = null, phy_schema = null, t_schema = null;
	static Document nt_schema_doc = null, log_anno = null, phy_anno = null;
	
	static AnnoValidator anno_validator;	
	
	static void parse_args(String[] args) throws Exception{
		for (int i = 0; i < args.length; i++){
			parser.parse(new InputSource(args[i]));
			
			Document temp_doc = parser.getDocument();
			String root_name = temp_doc.getDocumentElement().getNodeName();
			
			if (root_name.equals("temporalAnnotatedSchema")){
				log_schema = args[i];
				log_anno = temp_doc;
				
				continue;
			}
			
			if (root_name.equals("physicalAnnotatedSchema")){
				phy_schema = args[i];
				phy_anno = temp_doc;				
				
				continue;
			}
			
			if (temp_doc.getDocumentElement().getNodeName().equals("xs:schema")){
				nt_schema = args[i];
				
				parser.parse(nt_schema);
				nt_schema_doc = parser.getDocument();
				
				continue;
			}
			
			txml = args[i];
				
			doc = temp_doc;
	    	doc_root = temp_doc.getDocumentElement();	    	
		}
		
		/* try to get logical annotation's location from physical annotation */
		if ((log_anno == null) && (phy_anno != null)){
			NodeList nl = phy_anno.getElementsByTagName("temporalAnnotatations");
			
			if (nl.getLength() > 0){
				String location = ((Element)nl.item(0)).getAttribute("location");	
				
				parser.parse(new InputSource(location));
				log_anno = parser.getDocument();						
				log_schema = location;
				
				// try to get non-temporal schema's location from logical annotation
				if (nt_schema_doc == null){
					nl = log_anno.getElementsByTagName("snapshotSchema");
				
					if (nl.getLength() > 0){
						nt_schema = ((Element)nl.item(0)).getAttribute("location");
						
						parser.parse(nt_schema);
						nt_schema_doc = parser.getDocument();
					}
				}		
			}
		}
				
		/* try to get non-temporal schema's location from logical annotation */
		if ((nt_schema == null) && (log_anno != null)){
			NodeList nl = log_anno.getElementsByTagName("snapshotSchema");
			
			if (nl.getLength() > 0){
				nt_schema = ((Element)nl.item(0)).getAttribute("location");
				
				parser.parse(nt_schema);
				nt_schema_doc = parser.getDocument();
			}
		}
		
		if (txml == null){
			System.out.println("Temporal XML document is missing!");
			System.exit(0);
		}
		
		if (nt_schema == null){
			System.out.println("Non-temporal schema is missing!");
			System.exit(0);
		}
		
		return;		
	}
	
	public static void main(String args[]) throws Exception{
		if (args.length == 0){
			System.out.println("Usage: java TemporalValidator temporal-XML [non-temporal-schema | logical_annotation | physical_annotation]");	
			System.exit(0);
		}
		
		long begin_tm = System.currentTimeMillis();
		
		parser = new DOMParser();
		parser.setErrorHandler(new ErrHandler());
		
		parse_args(args);
		
		if ((log_anno != null) && (phy_anno != null)){
			if (!anno_validator.validLogPhyAnno(nt_schema_doc, log_anno, phy_anno)){
				System.out.println("Invalid logical or physical annotations!");
				System.exit(0);
			}	
		}
		else if (log_anno != null){
			if (!anno_validator.validLogicalAnno(nt_schema_doc, log_anno)){
				System.out.println("Invalid logical annotations!");
				System.exit(0);
			}	
		}
		else if (phy_anno != null){
			if (!anno_validator.validPhysicalAnno(nt_schema_doc, phy_anno)){
				System.out.println("Invalid physical annotations!");
				System.exit(0);
			}
		}
		
		t_schema = "temp_" + nt_schema.substring(nt_schema.lastIndexOf("/")+1);		
		
		anno_parser = new AnnoParser(log_anno, phy_anno);
		
		if (phy_schema != null){	// regular txml
			/* generate representational schema for regular txml */
			String[] mapper_args = new String[2];
			mapper_args[0] = nt_schema;
			mapper_args[1] = phy_schema;
			SchemaMapper.main(mapper_args);
			
			// reorder elements to make it consistent with schema
			String[] arrange_args = new String[3];
			arrange_args[0] = txml;
			arrange_args[1] = t_schema;
			arrange_args[2] = phy_schema;
			Arrange.main(arrange_args);
		}
		else{	// timevarying-root txml
			/* generate representational schema for timevarying-root txml */
			String[] mapper_args = new String[1];
			mapper_args[0] = nt_schema;
			SchemaMapper.main(mapper_args);
		}
	    
	    /* switch on the property of validaing upon schema */
	    parser.setFeature("http://xml.org/sax/features/validation", true);
        parser.setFeature("http://apache.org/xml/features/validation/schema", true);
        parser.setFeature("http://apache.org/xml/features/dom/include-ignorable-whitespace", false);
        
		parser.setProperty("http://apache.org/xml/properties/schema/external-noNamespaceSchemaLocation",
							t_schema);
		
		/* check validality of verids */
		validate_verid();
		
		if (log_anno != null)
			validate_nontv_parts();		/* check consistency of non-timevaring components */
		
		if (phy_schema == null)		// time-varying root TXML
			validate_tvroot_txml();		
		else						// regular TXML document									
			validate_regular_txml();
        
        long end_tm = System.currentTimeMillis();
        
        System.out.println("Valid document! The validation takes " + (end_tm - begin_tm) + " milliseconds.");
	}
	
	static void validate_regular_txml() throws Exception {		
		/* get top-level definitions in schema */
		get_ntschema_defs(nt_schema);
		
        /* create timestamp treewalker */
        ts_tw = ((DocumentTraversal)doc_root.getOwnerDocument()).createTreeWalker(doc_root,
														NodeFilter.SHOW_ELEMENT,
														new TS_TSFilterIn(),
														false);
									
		/* create timevarying attribute treewalker */
		tvattr_tw = ((DocumentTraversal)doc_root.getOwnerDocument()).createTreeWalker(doc_root,
														NodeFilter.SHOW_ELEMENT,
														new TS_TVAttrFilterIn(),
														false);	
        
        /* recursively validate elements from document root */
        validate(doc_root, sdf.parse(doc_root.getAttribute("cur_date")), sdf.parse(FOREVER));
		
		return;
	}
	
	static void validate_tvroot_txml() throws Exception {
        /* get all timestamp elements */
        NodeList ts_list = doc_root.getElementsByTagName("timestamp");
		
		ArrayList begin_list, end_list;
		begin_list = new ArrayList();
		end_list = new ArrayList();
		
		/* sort snapshots' timestamps */
		for (int i = 0; i < ts_list.getLength(); i++){
			Element ts1 = (Element)ts_list.item(i);
			
			Date ts_begin, ts_end;
			
			ts_begin = sdf.parse(ts1.getAttribute("vtbegin"));
			ts_end = sdf.parse(ts1.getAttribute("vtend"));
			
			if (ts_begin.compareTo(ts_end) >= 0){
				System.out.println("Error: vtend is before vtbegin!");
				System.exit(0);
			}
			
			int posi = -1;
			for (int j = 0; j < begin_list.size(); j++){
				Date t = (Date)begin_list.get(j);
								
				if (t.compareTo(ts_begin) > 0){
					posi = j;
					break;
				}								
			}
			
			if (posi < 0){
				begin_list.add(ts_begin);
				end_list.add(ts_end);
			}
			else{
				begin_list.add(posi, ts_begin);
				end_list.add(posi, ts_end);
			}
		}
		
		/* check overlapping of timestamps */
		for (int i = 0; i < begin_list.size() - 1; i++){
			Date e1, b2;
			
			e1 = (Date)end_list.get(i);
			b2 = (Date)begin_list.get(i + 1);
			
			if (e1.compareTo(b2) > 0){
				System.out.println("Error: timestamps overlap!");
				System.exit(0);
			}
		}		
		
		if (log_anno == null)
			return;
		
		/* 
		 *	compare the content of non-timevarying elements that
		 *	don't have a timevarying ancestor. Content of non-timevarying
		 *	elements that have timevarying ancestor are compared in
		 *	validate_nontv_parts()
		 */
		NodeList nl = doc_root.getChildNodes();
		
		ArrayList snapshot = new ArrayList();
		for (int i = 0; i < nl.getLength(); i++)
			if (nl.item(i) instanceof Element)
				snapshot.add(nl.item(i));				
		
		Element snap1 = (Element)snapshot.get(0);
		ArrayList c1 = get_non_tv_children(snap1);
		
		for (int i = 1; i < snapshot.size(); i++){
			Element snap2 = (Element)snapshot.get(i);
			
			ArrayList c2 = get_non_tv_children(snap2);
			
			if (c1.size() != c2.size()){
				System.out.println("Error: different size of non-timevarying children!");
				System.exit(0);
			}
				
			for (int j = 0; j < c1.size(); j++){
				Node n1, n2;
				
				n1 = (Node)c1.get(j);
				n2 = (Node)c2.get(j);
				
				if (!equal_nodes(n1, n2)){
					System.out.println("Error: non-timevarying nodes changes!");
					System.exit(0);
				}
			}
		
		}
		
		return;		
	}
	
	static void validate_verid() throws Exception{
		/*	elements with the same verid should have same name and at the same depth */		 
		
		Hashtable vid_name_table = new Hashtable();		// keep the name of elements having verid
		Hashtable vid_depth_table = new Hashtable();	// keep the depth of elements with verid
		
		/* keep the non-existence-varying elements grouped by their nearest ancestor that has timestamp */
		Hashtable non_ex_varying_elements = new Hashtable(); 
		
		TreeWalker tw1 = ((DocumentTraversal)doc).createTreeWalker(doc_root,
																	NodeFilter.SHOW_ELEMENT,
																	null,
																	false);
		
		/* keep element at next level */												
		ArrayList next_list = new ArrayList();
		
		/* keep element at current level */												
		ArrayList cur_list = new ArrayList();
		
		cur_list.add(doc_root);
		
		int depth = 0;
		
		/* traverse the dom tree using breadth first search */
		while (cur_list.size() > 0){
			for (int i = 0; i < cur_list.size(); i++){
				Element p = (Element)cur_list.get(i);
				
				tw1.setCurrentNode(p);
				
				Element c = (Element)tw1.firstChild();
				
				while (c != null){					
					String c_verid = c.getAttribute("verid");
					
					if (c_verid.length() > 0){
						/* check name consistency */
						String old_name = (String)vid_name_table.get(c_verid);						
						if (old_name != null){
							if (!c.getNodeName().equals(old_name)){
								System.out.println("Error: two different entities have the same verid " + c_verid + "!");
								System.exit(0);								
							}
						}
						else
							vid_name_table.put(c_verid, c.getNodeName());
						
						/* check depth consistency */
						Integer old_depth = (Integer)vid_depth_table.get(c_verid);						
						if (old_depth != null){
							if (depth != old_depth.intValue()){
								System.out.println("Error: elements with the same verid are not at the same depth.");
								System.exit(0);
							}
						}
						else
							vid_depth_table.put(c_verid, new Integer(depth));
							
						if (!anno_parser.isExistenceVarying(c) && (c.getAttribute("verid").length() > 0)){
							/* 
							 *	find the nearest ancestor that has timestamp,
							 *	c should be present in all the ancestor's period.
							 */
							
							Node ancestor = c.getParentNode();
							TreeWalker ts_tw = ((DocumentTraversal)doc).createTreeWalker(ancestor,
																						NodeFilter.SHOW_ELEMENT,
																						new TS_TSFilterIn(),
																						false);																						
							
							while (ancestor instanceof Element){
								ts_tw.setCurrentNode(ancestor);
								
								Element ancestor_ts = (Element)ts_tw.firstChild();
								
								if (ancestor_ts != null)
									break;
									
								ancestor = ancestor.getParentNode();
							}
							
							ArrayList non_ev_list = (ArrayList)non_ex_varying_elements.get(ancestor);
							
							if (non_ev_list == null)
								non_ev_list = new ArrayList();
								
							non_ev_list.add(c);
							
							non_ex_varying_elements.put(ancestor, non_ev_list);
						}
					}
					
					next_list.add(c);					
					
					tw1.setCurrentNode(c);
					
					c = (Element)tw1.nextSibling();
				}
			}
			
			cur_list = (ArrayList)next_list.clone();
			
			next_list.clear();
			
			depth++;
		}
		
		/* 
		 *	special processing for non-existence-varying element,
		 *	non-existence-varying elements must always be present
		 *	in the valid period of the nearest ancestor that has timestamp,
		 *	or if no such ancestor exists, non-existence-varying elements
		 *	must be present at all time
		 */		
		 
		TreeWalker ts_tw = ((DocumentTraversal)doc).createTreeWalker(doc_root,
																	NodeFilter.SHOW_ELEMENT,
																	new TS_TSFilterIn(),
																	false);
							
		
		Enumeration enum = non_ex_varying_elements.keys();
		
		while (enum.hasMoreElements()){
			Node ancestor = (Node)enum.nextElement();
			
			ArrayList non_ev_list = (ArrayList)non_ex_varying_elements.get(ancestor);
			
			Date ancestor_begin, ancestor_end;			
			if (ancestor instanceof Element){
				/* ancestor must have a timestamp */
				ts_tw.setCurrentNode(ancestor);
				Element ts = (Element)ts_tw.firstChild();
				
				ancestor_begin = sdf.parse(ts.getAttribute("vtbegin"));
				ancestor_end = sdf.parse(ts.getAttribute("vtend"));
			}
			else{
				/* ancestor must be document */
				ancestor_begin = sdf.parse(doc_root.getAttribute("cur_date"));
				ancestor_end = sdf.parse(FOREVER);
			}
			
			while (non_ev_list.size() > 0){
				Element e = (Element)non_ev_list.remove(0);
				
				ts_tw.setCurrentNode(e);
				Element ts = (Element)ts_tw.firstChild();
				
				if (ts == null)		/* the element is valid in ancestor's period if it has no timestamp */
					continue;
				
				/* 
				 *	get all elements in the list whose verids are the same as e,
				 *	the combination of their valid period should be equal to [ancestor_begin, ancestor_end)
				 */
				String verid = e.getAttribute("verid");
				
				ArrayList begin_list, end_list;
				
				begin_list = new ArrayList();
				end_list = new ArrayList();
				
				begin_list.add(sdf.parse(ts.getAttribute("vtbegin")));
				end_list.add(sdf.parse(ts.getAttribute("vtend")));
				
				int posi = 0;
				while (posi < non_ev_list.size()){
					e = (Element)non_ev_list.get(posi);
					
					if (e.getAttribute("verid").equals(verid)){
						non_ev_list.remove(posi);
						
						ts_tw.setCurrentNode(e);
						ts = (Element)ts_tw.firstChild();
						
						if (ts == null){
							System.out.println("Error: overlapping timestamps of elements (verid = " + verid + ")!");
							System.exit(0);
						}
						
						Date b1, e1;
						b1 = sdf.parse(ts.getAttribute("vtbegin"));
						e1 = sdf.parse(ts.getAttribute("vtend"));						
						
						int begin_posi = -1;
						for (int j = 0; j < begin_list.size(); j++)
							if (b1.before((Date)begin_list.get(j))){
								begin_posi = j;
								break;
							}
						
						if (begin_posi == -1){
							begin_list.add(b1);
							end_list.add(e1);	
						}
						else{
							begin_list.add(begin_posi, b1);
							end_list.add(begin_posi, e1);	
						}
					}
					else 
						posi++;
				}
				
				/* test if combination of timestamps is equal to the period of ancestor */
				if (ancestor_begin.compareTo(begin_list.get(0)) != 0){
					//System.out.println(ancestor_begin + " --- " + begin_list.get(0));
					System.out.println("Error: elements with verid " + verid + " violates existence-varying constraint.");
					System.exit(0);	
				}
				
				if (ancestor_end.compareTo(end_list.get(end_list.size()-1)) != 0){
					//System.out.println(ancestor_end + " --- " + end_list.get(end_list.size()-1));
					System.out.println("Error: elements with verid " + verid + " violates existence-varying constraint.");
					System.exit(0);	
				}
				
				for (int i = 1; i < begin_list.size(); i++){
					Date e1, b1;
					
					e1 = (Date)end_list.get(i-1);
					b1 = (Date)begin_list.get(i);
					
					if (!e1.equals(b1)){
						//System.out.println(b1 + " --- " + e1);
						System.out.println("Error: elements with verid " + verid + " violates existence-varying constraint.");
						System.exit(0);
					}
				}
			}
		}
		
		return;
	}
	
	static void validate(Element e, Date begin, Date end) throws Exception{
		Date new_begin = (Date)begin.clone();
		Date new_end = (Date)end.clone();
		
		if (anno_parser.isTimestampableElement(e)){
			/* get new valid period for decadents under this element */
			ts_tw.setCurrentNode(e);												

			Element ts = (Element)ts_tw.firstChild();
			
			if (ts != null){				
				new_begin = sdf.parse(ts.getAttribute("vtbegin"));
				new_end = sdf.parse(ts.getAttribute("vtend"));
			}	
		}		
		
		/* check timevarying attributes */
		validate_tv_attr(e, new_begin, new_end);
		
		/* get timestampable of children */
		ArrayList ts_children = get_ts_children(e);
		
		/* check validality of timestamps of timestampable children */
		validate_timestamps((ArrayList)ts_children.clone(), new_begin, new_end);
				
		/* check maxmimum and minimum occurs constrains given by non-temporal schema */
		validate_min_max_occurs((ArrayList)ts_children.clone(), new_begin, new_end);
		
		/* validate chil elements recursively */
		NodeList children = e.getChildNodes();			
		for (int i = 0; i < children.getLength(); i++){
			Node c = children.item(i);
			
			if (!(c instanceof Element))
				continue;
				
			if (c.getNodeName().equals("timestamp") ||
					c.getNodeName().equals("timevarying-attr"))
				continue;
			
			validate((Element)c, new_begin, new_end);
		}
		
		return;
	}
	
	/* find out required attributes of an element from non-temporal schema */
	static ArrayList get_req_attrs(Element parent){
		ArrayList req_attrs = new ArrayList();
		
		Element def = find_non_temporal_def(parent);
		
		if (def.getAttribute("ref").length() > 0)
			def = (Element)def_table.get(parent.getNodeName());		
		
		TreeWalker atw = ((DocumentTraversal)def.getOwnerDocument()).createTreeWalker(def,
														NodeFilter.SHOW_ELEMENT,
														null,
														false);		
														
		ArrayList nodes_past = new ArrayList();												
														
		Element e = (Element)atw.firstChild();
		while (true){
			if (e == null){
				if (nodes_past.size() == 0)
					break;
					
				e = (Element)nodes_past.remove(nodes_past.size()-1);
				
				atw.setCurrentNode(e);
				e = (Element)atw.nextSibling();
				
				continue;	
			}
			
			if (e.getNodeName().equals("xs:attribute")){
				if (e.getAttribute("use").equals("required"))
					req_attrs.add(e.getAttribute("name"));	
					
				atw.setCurrentNode(e);
				e = (Element)atw.nextSibling();
			}
			else if (e.getNodeName().equals("xs:element")){
				atw.setCurrentNode(e);
				e = (Element)atw.nextSibling();
			}
			else{
				nodes_past.add(e);
				
				atw.setCurrentNode(e);
				e = (Element)atw.firstChild();
			}
		}		
		
		return req_attrs;
	}
	
	/* test validality of timevarying attributes of the element */
	static void validate_tv_attr(Element e, Date begin, Date end) throws Exception{
		/* get required attributes of this element in non-temporal schema */
		ArrayList req_attrs = get_req_attrs(e);		
		
		NamedNodeMap attr_map = e.getAttributes();
		
		Hashtable tv_attrs = new Hashtable();
		
		tvattr_tw.setCurrentNode(e);
		
		Element ta = (Element)tvattr_tw.firstChild();
		while (ta != null){
			if (attr_map.getNamedItem(ta.getAttribute("name")) != null){
				// temporal and non-temporal attributes overlap in time
				System.out.println("Error: timestamps of timevarying attributes " + ta.getAttribute("name")
									+ " overlap!");
				System.exit(0);
			}
			
			// check range of timestamp
			Date ta_begin, ta_end;
			ta_begin = sdf.parse(ta.getAttribute("vtbegin"));
			ta_end = sdf.parse(ta.getAttribute("vtend"));
			
			if ((ta_begin.compareTo(begin) < 0) ||
					(ta_end.compareTo(end) > 0)){				
				System.out.println("Error: timestamp of timevarying attribute " + ta.getAttribute("name") +
									" exceeds range of its owner!");
				System.exit(0);
			}				
			
			// put timevaying attributes of the same name in one list
			// for checking timestamps later
			ArrayList tv_list = (ArrayList)tv_attrs.get(ta.getAttribute("name"));
			if (tv_list == null)
				tv_list = new ArrayList();
				
			tv_list.add(ta);
			
			tv_attrs.put(ta.getAttribute("name"), tv_list);
			
			tvattr_tw.setCurrentNode(ta);
			ta = (Element)tvattr_tw.nextSibling();
		}
		
		/* find required attributes in attribute map */
		for (int i = 0; i < attr_map.getLength(); i++)
			req_attrs.remove(attr_map.item(i).getNodeName());
		
		/* check timevarying attributes */
		Enumeration enum = tv_attrs.keys();
		while (enum.hasMoreElements()){
			// check if this attribute is a timevaying one in logical schema
			String attr_name = (String)enum.nextElement();
			if (!anno_parser.hasTVAttr(e, attr_name)){
				System.out.println("Error: illegal timevarying attribute " + attr_name);
				System.exit(0);
			}
			
			ArrayList ta_list = (ArrayList)tv_attrs.get(attr_name);
			
			// collect and test the timestamps of this attribute
			ArrayList begin_list, end_list;
			begin_list = new ArrayList();
			end_list = new ArrayList();			
			
			for (int i = 0; i < ta_list.size(); i++){
				Element ta1 = (Element)ta_list.get(i);
				
				Date ts_begin, ts_end;
				
				ts_begin = sdf.parse(ta1.getAttribute("vtbegin"));
				ts_end = sdf.parse(ta1.getAttribute("vtend"));
				
				if (ts_begin.compareTo(ts_end) >= 0){
					System.out.println("Error: invalid timestamp of attribute " + attr_name);
					System.exit(0);
				}
				
				int posi = -1;
				for (int j = 0; j < begin_list.size(); j++){
					Date t = (Date)begin_list.get(j);
					
					if (t.compareTo(ts_begin) > 0){
						posi = j;
						break;
					}								
				}
				
				if (posi < 0){
					begin_list.add(ts_begin);
					end_list.add(ts_end);
				}
				else{
					begin_list.add(posi, ts_begin);
					end_list.add(posi, ts_end);
				}
			}
			
			boolean required = false;
			
			// find if this timevarying attribute is required in non-temporal schema
			for (int i = 0; i < req_attrs.size(); i++){
				if (attr_name.equals(req_attrs.get(i))){
					required = true;
					
					req_attrs.remove(i);
					
					break;	
				}	
			}
			
			if (required == true){
				// the first vtbegin and the last vtend should be equal to
				// begin and end respectively
				
				Date first_b, last_e;
				
				first_b = (Date)begin_list.get(0);
				last_e = (Date)end_list.get(end_list.size()-1);
				
				if ((first_b.compareTo(begin) != 0) || (last_e.compareTo(end) != 0)){
					System.out.println("Error: required attribute " + attr_name + " is not present in all time!");
					System.exit(0);
				}
			}
			
			/*
			 *	check the timestamps of the this timevarying attribute: no overlap, 
			 *	no gap if the attribute is required
			 */
			for (int i = 0; i < begin_list.size() - 1; i++){
				Date e1, b2;
				
				e1 = (Date)end_list.get(i);
				b2 = (Date)begin_list.get(i + 1);
				
				if (required == true){
					if (e1.compareTo(b2) != 0){
						System.out.println("Error: required attribute " + attr_name + " is not present in all time!");
						System.exit(0);
					}
				}
				else{
					if (e1.compareTo(b2) > 0){
						System.out.println("Error: timestamps of attribute " + attr_name + " overlaps!");
						System.exit(0);
					}
				}
			}
		}
		
		/* some required attribute is missing */
		if (req_attrs.size() > 0){
			System.out.println("Error: required timevarying attributes are missing!");
			System.exit(0);
		}
		
		return;
	}
	
	static void validate_timestamps(ArrayList e_list, Date begin, Date end) throws Exception{
		/* check overlapping of timestamps */
		while (e_list.size() > 0){
			Element e = (Element)e_list.remove(0);
			
			String verid = e.getAttribute("verid");
			
			ts_tw.setCurrentNode(e);			
			Element ts = (Element)ts_tw.firstChild();
			
			if (ts == null){
				/*
				 *	no other elements in the list should have the same verid
				 *	because one is already found to be valid at all time
				 */
				for (int j = 0; j < e_list.size(); j++){
					Element e1 = (Element)e_list.get(j);
					
					if (e1.getAttribute("verid").equals(verid)){
						System.out.println("Error: timestamps of element " + e.getNodeName() +
											" (verid=" + e.getAttribute("verid") + ") overlap!");
						System.exit(0);
					}
				}	
			}
			else{
				ArrayList begin_list = new ArrayList();
				ArrayList end_list = new ArrayList();
				
				Date ts_b, ts_e;
				
				ts_b = sdf.parse(ts.getAttribute("vtbegin"));
				ts_e = sdf.parse(ts.getAttribute("vtend"));
				
				if (ts_e.getTime() <= ts_b.getTime()){
					System.out.println("Error: invalid timestamps!");
					System.exit(0);
				}
				
				begin_list.add(ts_b);
				end_list.add(ts_e);				
				
				int posi = 0;
								
				/* get timestamps of elements with the same verid and sort them */
				while (posi < e_list.size()){
					Element e1 = (Element)e_list.get(posi);
					
					if (!e1.getAttribute("verid").equals(verid)){
						posi++;
						continue;
					}
						
					e_list.remove(posi);
						
					ts_tw.setCurrentNode(e1);
					Element ts1 = (Element)ts_tw.firstChild();
					
					if (ts1 == null){
						System.out.println("Error: timestamps of element " + e.getNodeName() +
											" (verid=" + e.getAttribute("verid") + ") overlap!");
						System.exit(0);
					}
						
					Date begin1 = sdf.parse(ts1.getAttribute("vtbegin"));
					Date end1 = sdf.parse(ts1.getAttribute("vtend"));
					
					if (end1.getTime() <= begin1.getTime()){
						System.out.println("Error: invalid timestamps!");
						System.exit(0);
					}
					
					int k;
					for (k = 0; k < begin_list.size(); k++){												
						Date next_begin = (Date)begin_list.get(k);
						
						if (begin1.getTime() < next_begin.getTime())
							break;						
					}
					
					begin_list.add(k, begin1);
					end_list.add(k, end1);
				}
				
				Date first_begin, last_end;
				
				first_begin = (Date)begin_list.get(0);
				last_end = (Date)end_list.get(end_list.size() - 1);
				
				if ((first_begin.getTime() < begin.getTime()) || (last_end.getTime() > end.getTime())){
					System.out.println("Error: timestamps exceed range!");
					System.exit(0);
				}
				
				for (int j = 0; j < begin_list.size() - 1; j++){
					Date prev_end, next_begin;
					
					prev_end = (Date)end_list.get(j);
					next_begin = (Date)begin_list.get(j + 1);
					
					if (prev_end.getTime() > next_begin.getTime()){
						System.out.println("Error: timestamps of element " + e.getNodeName() +
											" (verid=" + e.getAttribute("verid") + ") overlap!");
						System.exit(0);
					}
				}
			}	
		}				
		
		return;	
	}
				
	static void validate_nontv_parts(){
		/* keep all elements having verid */
		Hashtable vid_elements = new Hashtable();				
		
		TreeWalker verid_tw = ((DocumentTraversal)doc).createTreeWalker(doc_root,
														NodeFilter.SHOW_ELEMENT,
														null,
														false);		
		
		/* traverse the DOM tree to get elements with verid */
		Element n = (Element)verid_tw.firstChild();		
		while (n != null){
			String verid = n.getAttribute("verid");
			
			if (verid.length() > 0){				
				ArrayList v_array = (ArrayList)vid_elements.get(verid);
				if (v_array == null)
					v_array = new ArrayList();
					
				v_array.add(n);
				
				vid_elements.put(verid, v_array);
			}
			
			verid_tw.setCurrentNode(n);
			n = (Element)verid_tw.nextNode();
		}
		
		
		Enumeration values = vid_elements.elements();		
		while (values.hasMoreElements()){
			ArrayList p_list = (ArrayList)values.nextElement();
			
			if (p_list.size() == 1)
				continue;			
		
			Element e1 = (Element)p_list.get(0);			
			ArrayList c1 = get_non_tv_children(e1);
			
			/* get non-timevarying attributes */
			NamedNodeMap map1 = e1.getAttributes();
			Hashtable non_tvattr = new Hashtable();
			for (int i = 0; i < map1.getLength(); i++){
				Attr attr = (Attr)map1.item(i);
				
				if (!anno_parser.hasTVAttr(e1, attr.getNodeName()))
					non_tvattr.put(attr.getNodeName(), attr.getNodeValue());
			}
			
			for (int i = 1; i < p_list.size(); i++){
				Element e2 = (Element)p_list.get(i);				
				
				/* compare values of non-timevarying attributes */
				NamedNodeMap map2 = e2.getAttributes();
				
				for (int j = 0; j < map2.getLength(); j++){
					Attr attr = (Attr)map2.item(j);	
					
					if (!anno_parser.hasTVAttr(e2, attr.getNodeName())){
						String v1, v2;
						
						v1 = (String)non_tvattr.get(attr.getNodeName());
						v2 = attr.getNodeValue();
						
						if (v1 != null){
							if (!v1.equals(v2)){
								System.out.println("Error: non-timevarying attribute " + 
													attr.getNodeName() + " changes!");
								System.exit(0);	
							}
						}
						else
							non_tvattr.put(attr.getNodeName(), attr.getNodeValue());
					}
				}

				ArrayList c2 = get_non_tv_children(e2);
				if (c1.size() != c2.size()){
					System.out.println("Error: unequal non-timevarying parts of element " + 
										e1.getNodeName() + " (verid=" + e1.getAttribute("verid") + ")!");
					System.exit(0);
				}				
					
				for (int j = 0; j < c1.size(); j++){
					Node n1, n2;
					
					n1 = (Node)c1.get(j);
					n2 = (Node)c2.get(j);
					
					if (!equal_nodes(n1, n2)){
						System.out.println("Error: unequal non-timevarying parts of element " + 
										e1.getNodeName() + " (verid=" + e1.getAttribute("verid") + ")!");
						System.exit(0);
					}				
				}				
			}
		}
		
		return;	
	}
	
	/* get top level definitions of non-temporal schema */
	static void get_ntschema_defs(String sch_file) throws Exception{
		DOMParser parser = new DOMParser();
		
		parser.parse(sch_file);
		Element sch_root = parser.getDocument().getDocumentElement();
		
		def_table = new Hashtable();
		
    	NodeList nl = sch_root.getChildNodes();
		
		for (int i = 0; i < nl.getLength(); i++){
			Node n = nl.item(i);
			
			if (!n.getNodeName().equals("xs:element"))
				continue;
				
			Element def = (Element)n;			
			
			def_table.put(def.getAttribute("name"), def);
		}
		
		return;
	}
	
	/* find the definition of element in non-temporal schema */
	static Element find_non_temporal_def(Element e){
		Element def;
		
		Element root_def = (Element)def_table.get(doc.getDocumentElement().getNodeName());
				
		if (e.equals(e.getOwnerDocument().getDocumentElement()))
			return root_def;
		
		// get the names of nodes on the path to e
		ArrayList node_name = new ArrayList();
		node_name.add(e.getNodeName());
		
		Node parent = e.getParentNode();
		while (parent instanceof Element){
			node_name.add(0, parent.getNodeName());
			parent = parent.getParentNode();
		}
		
		TreeWalker tw_def = ((DocumentTraversal)root_def.getOwnerDocument()).createTreeWalker(root_def,
														NodeFilter.SHOW_ELEMENT,
														null,
														false);
		ArrayList nodes_past = new ArrayList();
		
		int posi = 1;
		
		Node n = tw_def.firstChild();
		while (true){
			if (n == null){
				if (nodes_past.size() == 0){
					System.out.println("Error: can't find definition of " + e.getNodeName() +
										" in non-temporal schema!");
					System.exit(0);
				}
				
				n = (Node)nodes_past.remove(nodes_past.size()-1);
				
				tw_def.setCurrentNode(n);
				n = tw_def.nextSibling();	
				
				continue;
			}
			
			if (!n.getNodeName().equals("xs:element")){
				nodes_past.add(n);
				
				tw_def.setCurrentNode(n);
				n = tw_def.firstChild();
				
				continue;
			}
			
			def = (Element)n;
			
			if (def.getAttribute("name").equals(node_name.get(posi)) ||
					def.getAttribute("ref").equals(node_name.get(posi))){
				posi++;
				
				if (posi >= node_name.size())
					break;
					
				if (def.getAttribute("ref").length() > 0)
					def = (Element)def_table.get(def.getAttribute("ref"));
				
				tw_def.setCurrentNode(def);
				n = tw_def.firstChild();
			}
			else{
				tw_def.setCurrentNode(n);
				n = tw_def.nextSibling();	
			}
		}
		
		return def;
	}
	
	/* find the minimum occurs of element required by non-temporal schema */
	static int get_min_occur(Element c){
		Element def = find_non_temporal_def(c);
		
		String str_min_occur = def.getAttribute("minOccurs");
		
		if (str_min_occur.length() == 0)
			return 1;
		else
			return Integer.parseInt(str_min_occur);
	}
	
	/* find the maximum occurs of element required by non-temporal schema */
	static int get_max_occur(Element c){
		Element def = find_non_temporal_def(c);
		
		String str_max_occur = def.getAttribute("maxOccurs");
		
		if (str_max_occur.length() == 0)
			return 1;
		else if (str_max_occur.equals("unbounded"))
			return Integer.MAX_VALUE;
		else
			return Integer.parseInt(str_max_occur);
	}
	
	/* increase the number of occurs in the given period */	
	static void add_period(ArrayList begin_list, ArrayList end_list, Date begin, Date end, ArrayList occur_list){
		int begin_posi = 0, end_posi = 0;		
		
		for (int i = 0; i < end_list.size(); i++){
			Date list_end = (Date)end_list.get(i);
			
			if (begin.getTime() < list_end.getTime()){
				begin_posi = i;
				break;
			}
		}	
		
		for (int i = begin_posi; i < end_list.size(); i++){
			Date list_end = (Date)end_list.get(i);
			
			if (end.getTime() <= list_end.getTime()){
				end_posi = i;
				break;	
			}
		}
		
		if (begin_posi == end_posi){
			Date list_begin = (Date)begin_list.get(begin_posi);
			Date list_end = (Date)end_list.get(end_posi);
			
			if ((begin.getTime() == list_begin.getTime()) &&
					(end.getTime() == list_end.getTime())){
				int occur = ((Integer)occur_list.get(begin_posi)).intValue();
				
				occur++;
				
				occur_list.set(begin_posi, new Integer(occur));						
				
				return;
			}
			else if (begin.getTime() == list_begin.getTime()){				
				int old_occur = ((Integer)occur_list.get(begin_posi)).intValue();
				
				begin_list.add(begin_posi + 1, end.clone());
				end_list.add(begin_posi + 1, list_end.clone());
				occur_list.add(begin_posi + 1, new Integer(old_occur));
				
				end_list.set(begin_posi, end);
				occur_list.set(begin_posi, new Integer(old_occur + 1));
				
				return;
			}
			else if (end.getTime() == list_end.getTime()){				
				int old_occur = ((Integer)occur_list.get(begin_posi)).intValue();				
				
				end_list.set(begin_posi, begin.clone());				
				
				begin_list.add(begin_posi + 1, begin.clone());
				end_list.add(begin_posi + 1, list_end.clone());
				occur_list.add(begin_posi + 1, new Integer(old_occur + 1));
				
				return;
			}
			else{
				int old_occur = ((Integer)occur_list.get(begin_posi)).intValue();				
				
				end_list.set(begin_posi, begin.clone());
				
				begin_list.add(begin_posi + 1, begin.clone());
				end_list.add(begin_posi + 1, end.clone());
				occur_list.add(begin_posi + 1, new Integer(old_occur + 1));
				
				begin_list.add(begin_posi + 2, end.clone());
				end_list.add(begin_posi + 2, list_end.clone());
				occur_list.add(begin_posi + 2, new Integer(old_occur));
				
				return;
			}
		}
		else{
			Date list_begin = (Date)begin_list.get(begin_posi);
			Date list_end = (Date)end_list.get(end_posi);
			
			for (int i = begin_posi + 1; i <= end_posi - 1; i++){
				int occur = ((Integer)occur_list.get(i)).intValue();
				
				occur++;
				
				occur_list.set(i, new Integer(occur));
			}			
			
			if (list_begin.getTime() == begin.getTime()){
				int occur = ((Integer)occur_list.get(begin_posi)).intValue();
				
				occur++;
				
				occur_list.set(begin_posi, new Integer(occur));						
			}
			else{
				int old_occur = ((Integer)occur_list.get(begin_posi)).intValue();
				
				Date old_end = (Date)end_list.get(begin_posi);
				
				end_list.set(begin_posi, begin.clone());				
				
				begin_list.add(begin_posi + 1, begin.clone());
				end_list.add(begin_posi + 1, old_end.clone());
				occur_list.add(begin_posi + 1, new Integer(old_occur + 1));
			}
			
			if (list_end.getTime() == end.getTime()){
				int occur = ((Integer)occur_list.get(end_posi)).intValue();
				
				occur++;
				
				occur_list.set(end_posi, new Integer(occur));						
			}
			else{
				int old_occur = ((Integer)occur_list.get(end_posi)).intValue();
				
				end_list.set(end_posi, end.clone());
				occur_list.set(end_posi, new Integer(old_occur + 1));
				
				begin_list.add(end_posi + 1, end.clone());
				end_list.add(end_posi + 1, list_end.clone());
				occur_list.add(end_posi + 1, new Integer(old_occur));
				
			}
		}
	}
	
	/* check the minimum and maximum constraints specified in non-temporal schema in all periods */
	static void validate_min_max_occurs(ArrayList children, Date begin, Date end) throws Exception{
		while (children.size() > 0){
			Element c = (Element)children.remove(0);
			
			int min_occur = get_min_occur(c);
			int max_occur = get_max_occur(c);
			
			if ((min_occur > 0) || (max_occur < Integer.MAX_VALUE)){
				ArrayList begin_list = new ArrayList();	// begin of each period
				ArrayList end_list = new ArrayList();	// end of each period
				
				ArrayList occur_list = new ArrayList();	// number of occurs in each period
				
				begin_list.add(begin);
				end_list.add(end);
				
				occur_list.add(new Integer(0));
				
				ts_tw.setCurrentNode(c);
				Element ts = (Element)ts_tw.firstChild();
				
				Date c_begin, c_end;
				if (ts != null){
					c_begin = sdf.parse(ts.getAttribute("vtbegin"));
					c_end = sdf.parse(ts.getAttribute("vtend"));
				}
				else{
					c_begin = (Date)begin.clone();
					c_end = (Date)end.clone();
				}					
				
				add_period(begin_list, end_list, c_begin, c_end, occur_list);
				
				int posi = 0;
				while (posi < children.size()){
					Element	c1 = (Element)children.get(posi);
					
					if (c1.getNodeName().equals(c.getNodeName())){
						children.remove(posi);
						
						ts_tw.setCurrentNode(c1);
						ts = (Element)ts_tw.firstChild();
						
						if (ts != null){
							c_begin = sdf.parse(ts.getAttribute("vtbegin"));
							c_end = sdf.parse(ts.getAttribute("vtend"));
						}
						else{
							c_begin = (Date)begin.clone();
							c_end = (Date)end.clone();
						}					
						
						add_period(begin_list, end_list, c_begin, c_end, occur_list);
					}
					else
						posi++;
				}
				
				for (int i = 0; i < occur_list.size(); i++){
					int occur = ((Integer)occur_list.get(i)).intValue();
					
					if ((min_occur > 0) && (occur < min_occur)){
						System.out.println("Error: element " + c.getNodeName() + 
											" (verid=" + c.getAttribute("verid") + 
											") must occur at least " + min_occur + " times in all periods!");
						System.exit(0);
					}
						
					if ((max_occur < Integer.MAX_VALUE) && (occur > max_occur)){
						System.out.println("Error: element " + c.getNodeName() + 
											" (verid=" + c.getAttribute("verid") + 
											") can't occur more than " + max_occur + " times in all periods!");
						System.exit(0);					
					}
				}
			}
			else{
				// remove children from the list that don't have min/max occurs requirement
				int posi = 0;
				while (posi < children.size()){
					Element	c1 = (Element)children.get(posi);
					
					if (c1.getNodeName().equals(c.getNodeName())){
						children.remove(posi);	
					}
					else
						posi++;
				}
			}
		}
		
		return;	
	}
	
	/* recursively compare non-timevarying parts of two entities */
	static boolean equal_nodes(Node n1, Node n2){
		if (n1.getNodeType() != n2.getNodeType())
			return false;
			
		if (!n1.getNodeName().equals(n2.getNodeName()))
			return false;
			
		if (!(n1 instanceof Element)){
			String v1 = n1.getNodeValue();
			String v2 = n2.getNodeValue();
			
			if ((v1 != null) && (!v1.equals(v2))){
				// System.out.println(v1 + " : " + v2);
				
				return false;
			}
			
			if ((v1 == null) && (v2 != null)){
				// System.out.println(v1 + " : " + v2);
				return false;
			}
				
			return true;		
		}
		
		Element e1 = (Element)n1;
		Element e2 = (Element)n2;		
		
		ArrayList children1 = get_non_tv_children(e1);
		ArrayList children2 = get_non_tv_children(e2);
		
		if (children1.size() != children2.size())
			return false;
				
		for (int i = 0; i < children1.size(); i++){
			Node c1 = (Node)children1.get(i);
			Node c2 = (Node)children2.get(i);
			
			if (equal_nodes(c1, c2) == false)				
				return false;
		}		
		
		return true;
	}
	
	/* return non-timevarying components of the element */
	static ArrayList get_non_tv_children(Element parent){
		ArrayList non_tv_children = new ArrayList();
		
		NodeList nl = parent.getChildNodes();
		
		for (int i = 0; i < nl.getLength(); i++){
			Node n = nl.item(i);
			
			if (!(n instanceof Element)){
				if (parent.getAttribute("verid").length() == 0)
					non_tv_children.add(n);
					
				continue;
			}			
			
			if (n.getNodeName().equals("timestamp") || 
					n.getNodeName().equals("time-varyingattr"))
				continue;
				
			if (((Element)n).getAttribute("verid").length() > 0)
				continue;
			
			non_tv_children.add(n);
		}
		
		return non_tv_children;
	}
	
	/* get timestampable children of the element */
	static ArrayList get_ts_children(Element parent){
		ArrayList ts_children = new ArrayList();
		
		NodeList nl = parent.getChildNodes();
		
		for (int i = 0; i < nl.getLength(); i++){
			Node n = nl.item(i);
			
			if (!(n instanceof Element))
				continue;
				
			Element e = (Element)n;
			
			if (anno_parser.isTimestampableElement(e))
    			ts_children.add(e);
		}
		
		return ts_children;
	}	
}

class ErrHandler implements ErrorHandler
{
	public void error(SAXParseException exception){
		System.out.println("\nerror: " + exception);
	}
     
 	public void fatalError(SAXParseException exception){
 		System.out.println("\nfatal error: " + exception);
 	}
 	
    public void warning(SAXParseException exception){
    	System.out.println("\nwarning: " + exception);
    }	
}
