/*
 * File:	Unsquash.java
 * author:	Haitao Liu (haitaol@cs.arizona.edu)
 *
 * The tauXSchema toolset, Version 1.0
 *
 * Copyright (c) 2002-2003 
 * The Department of Computer Science of the University of Arizona.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by 
 *       the Department of Computer Science of the University of Arizona."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear. 
 *
 * 4. The names "tauXSchema" or "TAU (Temporal Access for Users)" 
 *    must not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact tau@cs.arizona.edu.
 *
 * 5. Products derived from this software may not be called "tauXSchema",
 *    nor may "tauXSchema" or "TAU (Temporal Access for Users)" appear in 
 *    their name, without prior written permission of the Department of 
 *    Computer Science of the University of Arizona.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE Department of Computer Science of 
 * the University of Arizona OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This toolset contains useful tools to manage temporal XML document.
 * For more information about tXSchema and temporal XML document, 
 * visit the homepage of TAU Project <http://www.cs.arizona.edu/tau/>.
 */

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.apache.xpath.*;
import java.util.*;
import org.apache.xerces.dom3.*;
import org.apache.xerces.parsers.*;
import org.w3c.dom.ls.*;
import java.io.*;
import org.w3c.dom.traversal.*;
import org.apache.xerces.parsers.*;
import org.xml.sax.*;
import java.text.*;

public class Unsquash
{	
	static Document doc;
	
	public static void main(String[] args) throws Exception {
		if (args.length == 0){
			System.out.println("Usage: java Unsquash temporal-document");
			System.exit(0);
		}
		
		DOMParser parser = new DOMParser();

        System.setProperty(DOMImplementationRegistry.PROPERTY,
    		"org.apache.xerces.dom.DOMImplementationSourceImpl");
		DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();

		DOMImplementationLS impl =
			(DOMImplementationLS)registry.getDOMImplementation("LS-Load");

		DOMWriter writer = impl.createDOMWriter();

		parser.parse(new InputSource(args[0]));
		doc = parser.getDocument();
		
		/* collect the beginning points of all periods */
		ArrayList instants = new ArrayList();
		
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
		
		NodeList ts_list = doc.getElementsByTagName("timestamp");
		for (int i = 0; i < ts_list.getLength(); i++){
			Element ts = (Element)ts_list.item(i);
			
			String begin = ts.getAttribute("vtbegin");			
			
			int j;
			for (j = 0; j < instants.size(); j++){
				String b1 = (String)instants.get(j);
				
				if (begin.equals(b1))
					break;
				
				if (sdf.parse(begin).before(sdf.parse(b1))){
					instants.add(j, begin);
					break;
				}
			}
			
			if (j >= instants.size())
				instants.add(begin);
		}
		
		NodeList ta_list = doc.getElementsByTagName("timevarying-attr");
		for (int i = 0; i < ta_list.getLength(); i++){
			Element ta = (Element)ta_list.item(i);
			
			String begin = ta.getAttribute("vtbegin");
			
			int j;
			for (j = 0; j < instants.size(); j++){
				String b1 = (String)instants.get(j);
				
				if (begin.equals(b1))
					break;
				
				if (sdf.parse(begin).before(sdf.parse(b1))){
					instants.add(j, begin);
					break;
				}
			}
			
			if (j >= instants.size())
				instants.add(begin);
		}
		
		for (int i = 0; i < instants.size(); i++){
			TDocumentImpl tdoc = new TDocumentImpl(doc);
			
			tdoc.setModeCurrent("yyyy-MM-dd", (String)instants.get(i));			
			
			Element snap_root = cloneElement(tdoc.getDocumentElement());
			
			snap_root.setAttribute("cur_date", (String)instants.get(i));
			
			System.out.println("Write out snapshot " + i + ": " + (String)instants.get(i));
			writer.writeNode(new FileOutputStream("snapshot" + i + ".xml"), snap_root);
		}
	}	
	
	static Element cloneElement(Element e){
		Element e_clone = doc.createElement(e.getNodeName());
		
		NodeList nl = e.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++){
			Node n = nl.item(i);
			
			if (!(n instanceof Element)){				
				e_clone.appendChild(n.cloneNode(true));
			}
			else{
				Element child_e = cloneElement((Element)n);
				e_clone.appendChild(child_e);	
			}
		}
		
		NamedNodeMap map = e.getAttributes();
		for (int i = 0; i < map.getLength(); i++){
			Attr attr = (Attr)map.item(i);	
			
			e_clone.setAttribute(attr.getNodeName(), attr.getNodeValue());
		}
		
		return e_clone;
	}
}